@@ -3,15 +3,15 @@
 (*    Copyright 2014 OCamlPro                                             *)
 (*    Copyright 2012 INRIA                                                *)
 (*                                                                        *)
-(*  All rights reserved.This file is distributed under the terms of the   *)
-(*  GNU Lesser General Public License version 3.0 with linking            *)
-(*  exception.                                                            *)
-(*                                                                        *)
-(*  OPAM is distributed in the hope that it will be useful, but WITHOUT   *)
-(*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY    *)
-(*  or FITNESS FOR A PARTICULAR PURPOSE.See the GNU General Public        *)
-(*  License for more details.                                             *)
+(*  All rights reserved. This file is distributed under the terms of the  *)
+(*  GNU Lesser General Public License version 2.1, with the special       *)
+(*  exception on linking described in the file LICENSE.                   *)
 (*                                                                        *)
 (**************************************************************************)
 
+(** (generated) Current git version of OPAM *)
+
+(** This is defined only at the client lib level to avoid triggering
+    full recompilations all the time *)
+
 val version: string option
